﻿using System;
using System.Collections.Generic;
using System.Data.SqlTypes;
using System.Text;
using System.Linq;
using System.Xml.Linq;
using Microsoft.SqlServer.Types;
using System.Diagnostics;

namespace DataConDemoWebRole.Business
{
    /// <summary>
    /// This class contains conversion methods for going from SqlGeography types
    /// to GeoJson types and vice versa.
    /// </summary>
    public static class GeoJsonConvert
    {
        public static readonly int SRID = 4326;

        /// <summary>
        /// Build an SqlGeography Point from a GeoJson Point.
        /// </summary>
        /// <param name="point">GeoJson Point</param>
        /// <returns>SqlGeography Point</returns>
        public static SqlGeography GeographyFromGeoJsonPoint(Models.GeoJson.Point point)
        {
            var geob = new SqlGeographyBuilder();
            geob.SetSrid(SRID);
            geob.BeginGeography(OpenGisGeographyType.Point);
            geob.BeginFigure(point.coordinates[0], point.coordinates[1]);
            geob.EndFigure();
            geob.EndGeography();
            var geog = geob.ConstructedGeography;
            Debug.WriteLine(geog.AsGml().Value);
            return geog;
        }


        /// <summary>
        /// Build an SqlGeography Polygon from a GeoJson Polygon.
        /// </summary>
        /// <param name="poly">GeoJson Polygon</param>
        /// <returns>SqlGeography Polygon</returns>
        public static SqlGeography GeographyFromGeoJsonPolygon(Models.GeoJson.Polygon poly)
        {
            var geob = new SqlGeographyBuilder();
            geob.SetSrid(SRID);
            geob.BeginGeography(OpenGisGeographyType.Polygon);
            geob.BeginFigure(poly.coordinates[0][0][0], poly.coordinates[0][0][1]);
            foreach (var pair in poly.coordinates[0].Skip(1))
            {
                geob.AddLine(pair[0], pair[1]);
            }
            geob.EndFigure();
            geob.EndGeography();
            var geog = geob.ConstructedGeography;
            Debug.WriteLine(geog.AsGml().Value);
            return geog;
        }

        /// <summary>
        /// Build an SqlGeography Polygon from a bounding box.
        /// </summary>
        /// <param name="box">Bounding Box</param>
        /// <returns>SqlGeography Polygon</returns>
        public static SqlGeography GeographyFromBoundingBox(Models.BoundingBox box)
        {
            var geob = new SqlGeographyBuilder();
            geob.SetSrid(SRID);
            geob.BeginGeography(OpenGisGeographyType.Polygon);
            geob.BeginFigure(box.North, box.West);
            geob.AddLine(box.South, box.West);
            geob.AddLine(box.South, box.East);
            geob.AddLine(box.North, box.East);
            geob.AddLine(box.North, box.West);
            geob.EndFigure();
            geob.EndGeography();
            var geog = geob.ConstructedGeography;
            Debug.WriteLine(geog.AsGml().Value);
            return geog;
        }

        /// <summary>
        /// Build an SqlGeography Polygon from a bounding box given in discrete coordinates.
        /// </summary>
        /// <param name="nwLat">Northwest Latitude</param>
        /// <param name="nwLong">Northwest Longitude</param>
        /// <param name="seLat">Southeast Latitude</param>
        /// <param name="seLong">Southeast Longitude</param>
        /// <returns>SqlGeography Polygon</returns>
        public static SqlGeography GeograhyFromBoundingBoxNwSe(double nwLat, double nwLong, double seLat, double seLong)
        {
            return GeographyFromBoundingBox(new Models.BoundingBox()
            {
                East = seLong,
                North = nwLat,
                South = seLat,
                West = nwLong
            });
        }

        /// <summary>
        /// Build an SqlGeography LineString from a GeoJson LineString.
        /// </summary>
        /// <param name="line">GeoJson LineString</param>
        /// <returns>SqlGeography LineString</returns>
        public static SqlGeography GeographyFromGeoJsonLineString(Models.GeoJson.LineString line)
        {
            var geob = new SqlGeographyBuilder();
            geob.SetSrid(SRID);
            geob.BeginGeography(OpenGisGeographyType.LineString);
            geob.BeginFigure(line.coordinates[0][0], line.coordinates[0][1]);
            foreach (var pair in line.coordinates.Skip(1))
            {
                geob.AddLine(pair[0], pair[1]);
            }
            geob.EndFigure();
            geob.EndGeography();
            var geog = geob.ConstructedGeography;
            Debug.WriteLine(geog.AsGml().Value);
            return geog;
        }


        private static void CheckGeographyType(SqlGeography geography, params string[] permitted)
        {
            string stype = (string)geography.STGeometryType();
            if (!permitted.Contains(stype))
            {
                throw new ArgumentException("This conversion cannot handle " + stype + ", it can only handle the following types: " + string.Join(",", permitted));
            }
        }

        /// <summary>
        /// Build a GeoJson Point from an SqlGeography Point.
        /// </summary>
        /// <param name="geography"SqlGeography Point></param>
        /// <returns>GeoJson Point</returns>
        public static Models.GeoJson.Point GeoJsonPointFromGeography(SqlGeography geography)
        {
            CheckGeographyType(geography, "Point");

            var point = new Models.GeoJson.Point()
            {
                coordinates = new double[2]
            };
            //we can safely round the lat/long to 5 decimal places as thats 1.11m at equator, reduces data transfered to client
            point.coordinates[0] = Math.Round((double)geography.Lat, 5);
            point.coordinates[1] = Math.Round((double)geography.Long, 5);
            return point;
        }

        /// <summary>
        ///  Build a GeoJson LineString from an SqlGeography LineString.
        /// </summary>
        /// <param name="geography">SqlGeography LineString</param>
        /// <returns>GeoJson LineString</returns>
        public static Models.GeoJson.LineString GeoJsonLineStringFromGeography(SqlGeography geography)
        {
            CheckGeographyType(geography, "LineString");

            var points = new List<double[]>();
            for (int i = 1; i <= geography.STNumPoints(); i++)
            {
                var point = new double[2];
                var sp = geography.STPointN(i);
                //we can safely round the lat/long to 5 decimal places as thats 1.11m at equator, reduces data transfered to client
                point[0] = Math.Round((double)sp.Lat, 5);
                point[1] = Math.Round((double)sp.Long, 5);
                points.Add(point);
            }
            return new Models.GeoJson.LineString()
            {
                coordinates = points.ToArray()
            };
        }

        /// <summary>
        ///  Build a GeoJson LineString from an SqlGeography LineString.
        /// </summary>
        /// <param name="geography">SqlGeography LineString or MultiLineString</param>
        /// <returns>GeoJson MultiLineString</returns>
        public static Models.GeoJson.MultiLineString GeoJsonMultiLineStringFromGeography(SqlGeography geography)
        {
            if ((string)geography.STGeometryType() == "LineString")
            {
                var points = new List<double[]>();
                for (int i = 1; i <= geography.STNumPoints(); i++)
                {
                    var point = new double[2];
                    var sp = geography.STPointN(i);
                    //we can safely round the lat/long to 5 decimal places as thats 1.11m at equator, reduces data transfered to client
                    point[0] = Math.Round((double)sp.Lat, 5);
                    point[1] = Math.Round((double)sp.Long, 5);
                    points.Add(point);
                }
                var mls = new Models.GeoJson.MultiLineString()
                {
                    coordinates = new double[1][][]
                };
                mls.coordinates[0] = points.ToArray();
                return mls;
            }

            if ((string)geography.STGeometryType() == "MultiLineString")
            {
                var lscoords = new List<double[][]>();
                for (var g = 1; g <= geography.STNumGeometries(); g++)
                {
                    var geo = geography.STGeometryN(g);
                    Debug.Assert((string)geo.STGeometryType() == "LineString");
                    var points = new List<double[]>();

                    for (int i = 1; i <= geo.STNumPoints(); i++)
                    {
                        var point = new double[2];
                        var sp = geography.STPointN(i);
                        //we can safely round the lat/long to 5 decimal places as thats 1.11m at equator, reduces data transfered to client
                        point[0] = Math.Round((double)sp.Lat, 5);
                        point[1] = Math.Round((double)sp.Long, 5);
                        points.Add(point);
                    }
                    lscoords.Add(points.ToArray());
                }
                var mls = new Models.GeoJson.MultiLineString()
                {
                    coordinates = lscoords.ToArray()
                };
                return mls;
            }

            throw new ArgumentException("GeoJsonMultiLineStringFromGeography will only convert MultiLineString and LineString geometries.");
        }


        private static double[][][] PolygonCoordinatesFromGeography(SqlGeography polygon)
        {
            Debug.Assert((string)polygon.STGeometryType() == "Polygon");

            // convert only the first (exterior) ring
            var ring = polygon.STGeometryN(1);
            var points = new List<double[]>();

            for (int i = 1; i <= ring.STNumPoints(); i++)
            {
                var point = new double[2];
                var sp = ring.STPointN(i);
                //we can safely round the lat/long to 5 decimal places as thats 1.11m at equator, reduces data transfered to client
                point[0] = Math.Round((double)sp.Lat, 5);
                point[1] = Math.Round((double)sp.Long, 5);
                points.Add(point);
            }

            var coordinates = new double[1][][];
            coordinates[0] = points.ToArray();
            return coordinates;
        }

        /// <summary>
        /// Create a GeoJson MultiPolygon from an SQL Server geography type.
        /// States and Countries are MultiPolygons on the client since they can come in many pieces.
        /// In the database, the geography may be a MultiPolygon or Polygon.
        /// Due to reduction, the geography may be a GeometryCollection containing 
        /// any types. This method extracts all external polygons from a geography
        /// and constructs a GeoJsonMultiPolygon.
        /// 
        /// To keep things simple in the client, we treat a Polygon as a Multipolygon
        /// containing a single Polygon.
        /// </summary>
        /// <remarks>
        /// Indices in a MultiPolygon are:
        /// 1. polygon 
        /// 2. ring (only first ring is used since only exterior is drawn)
        /// 3. point
        /// 4. lat, long
        /// </remarks>
        /// <param name="geography">geography</param>
        /// <returns>MultiPolygon</returns>
        public static Models.GeoJson.MultiPolygon GeoJsonMultiPolygonFromGeography(SqlGeography geography)
        {
            var geoType = (string)geography.STGeometryType();
            var polycoords = new List<double[][][]>();
            if (geoType == "Polygon")
            {
                polycoords.Add(PolygonCoordinatesFromGeography(geography));
            }

            // For a MultiPolygon or GeometryCollection, the same is repeated for each contained polygon
            else if (geoType == "MultiPolygon" || geoType == "GeometryCollection")
            {
                for (var g = 1; g <= geography.STNumGeometries(); g++)
                {
                    var geo = geography.STGeometryN(g);
                    if ((string)geo.STGeometryType() == "Polygon")
                    {
                        polycoords.Add(PolygonCoordinatesFromGeography(geo));
                    }
                    else
                    {
                        Debug.WriteLine("Ignoring " + (string)geo.STGeometryType());
                    }
                }
            }
            else
            {
                throw new ArgumentException("GeoJsonMultiPolygonFromGeography will only convert GeometryCollection, MultiPolygon and Polygon geometries.");
            }

            // return a MultiPolygon containing the coordinates of the polygons
            var mpoly = new Models.GeoJson.MultiPolygon()
            {
                coordinates = polycoords.ToArray()
            };
            return mpoly;
        }

    }
}